@@ -14,18 +14,19 @@
 
 #![allow(clippy::derive_partial_eq_without_eq)]
 
-use crate::{ei, util, Interface, Object, ParseError, PendingRequestResult};
+use crate::{ei, handshake::HandshakeResp, util, Error, Interface, Object, PendingRequestResult};
 use std::{
     collections::{HashMap, VecDeque},
     fmt, io,
     os::unix::io::OwnedFd,
-    sync::Arc,
+    sync::{
+        atomic::{AtomicU32, Ordering},
+        Arc,
+    },
 };
 
-// struct ReceiverStream(EiEventStream, ());
-
 #[derive(Debug)]
-pub enum Error {
+pub enum EventError {
     DeviceEventBeforeDone,
     DeviceSetupEventAfterDone,
     SeatSetupEventAfterDone,
@@ -34,7 +35,7 @@ pub enum Error {
     UnexpectedHandshakeEvent,
 }
 
-impl fmt::Display for Error {
+impl fmt::Display for EventError {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self {
             Self::DeviceEventBeforeDone => write!(f, "device event before done"),
@@ -47,7 +48,35 @@ impl fmt::Display for Error {
     }
 }
 
-impl std::error::Error for Error {}
+impl std::error::Error for EventError {}
+
+#[derive(Debug)]
+struct ConnectionInner {
+    context: ei::Context,
+    handshake_resp: HandshakeResp,
+    serial: AtomicU32,
+}
+
+#[derive(Clone, Debug)]
+pub struct Connection(Arc<ConnectionInner>);
+
+impl Connection {
+    pub fn connection(&self) -> &ei::Connection {
+        &self.0.handshake_resp.connection
+    }
+
+    pub fn flush(&self) -> rustix::io::Result<()> {
+        self.0.context.flush()
+    }
+
+    fn update_serial(&self, value: u32) {
+        self.0.serial.store(value, Ordering::Relaxed);
+    }
+
+    pub fn serial(&self) -> u32 {
+        self.0.serial.load(Ordering::Relaxed)
+    }
+}
 
 pub struct EiEventConverter {
     pending_seats: HashMap<ei::Seat, SeatInner>,
@@ -57,11 +86,12 @@ pub struct EiEventConverter {
     device_for_interface: HashMap<Object, Device>,
     events: VecDeque<EiEvent>,
     pending_events: VecDeque<EiEvent>,
-    serial: u32,
+    callbacks: HashMap<ei::Callback, Box<dyn FnOnce(u64)>>,
+    connection: Connection,
 }
 
 impl EiEventConverter {
-    pub fn new(serial: u32) -> Self {
+    pub fn new(context: &ei::Context, handshake_resp: HandshakeResp) -> Self {
         Self {
             pending_seats: HashMap::new(),
             seats: HashMap::new(),
@@ -70,10 +100,19 @@ impl EiEventConverter {
             device_for_interface: HashMap::new(),
             events: VecDeque::new(),
             pending_events: VecDeque::new(),
-            serial,
+            callbacks: HashMap::new(),
+            connection: Connection(Arc::new(ConnectionInner {
+                context: context.clone(),
+                serial: AtomicU32::new(handshake_resp.serial),
+                handshake_resp,
+            })),
         }
     }
 
+    pub fn connection(&self) -> &Connection {
+        &self.connection
+    }
+
     // Based on behavior of `queue_event` in libei
     fn queue_event(&mut self, mut event: EiEvent) {
         if event.time_mut().is_some() {
@@ -93,10 +132,10 @@ impl EiEventConverter {
         }
     }
 
-    pub fn handle_event(&mut self, event: ei::Event) -> Result<(), Error> {
+    pub fn handle_event(&mut self, event: ei::Event) -> Result<(), EventError> {
         match event {
             ei::Event::Handshake(_handshake, _event) => {
-                return Err(Error::UnexpectedHandshakeEvent);
+                return Err(EventError::UnexpectedHandshakeEvent);
             }
             ei::Event::Connection(_connection, event) => match event {
                 ei::connection::Event::Seat { seat } => {
@@ -134,9 +173,11 @@ impl EiEventConverter {
                     // TODO
                 }
             },
-            ei::Event::Callback(_callback, event) => match event {
-                ei::callback::Event::Done { callback_data: _ } => {
-                    // TODO
+            ei::Event::Callback(callback, event) => match event {
+                ei::callback::Event::Done { callback_data } => {
+                    if let Some(cb) = self.callbacks.remove(&callback) {
+                        cb(callback_data);
+                    }
                 }
             },
             ei::Event::Pingpong(_ping_pong, event) => match event {},
@@ -145,21 +186,21 @@ impl EiEventConverter {
                     let seat = self
                         .pending_seats
                         .get_mut(&seat)
-                        .ok_or(Error::SeatSetupEventAfterDone)?;
+                        .ok_or(EventError::SeatSetupEventAfterDone)?;
                     seat.name = Some(name);
                 }
                 ei::seat::Event::Capability { mask, interface } => {
                     let seat = self
                         .pending_seats
                         .get_mut(&seat)
-                        .ok_or(Error::SeatSetupEventAfterDone)?;
+                        .ok_or(EventError::SeatSetupEventAfterDone)?;
                     seat.capabilities.insert(interface, mask);
                 }
                 ei::seat::Event::Done => {
                     let seat = self
                         .pending_seats
                         .remove(&seat)
-                        .ok_or(Error::SeatSetupEventAfterDone)?;
+                        .ok_or(EventError::SeatSetupEventAfterDone)?;
                     let seat = Seat(Arc::new(seat));
                     self.seats.insert(seat.0.seat.clone(), seat.clone());
                     self.queue_event(EiEvent::SeatAdded(SeatAdded { seat }));
@@ -168,7 +209,7 @@ impl EiEventConverter {
                     let seat = self
                         .seats
                         .get_mut(&seat)
-                        .ok_or(Error::SeatEventBeforeDone)?;
+                        .ok_or(EventError::SeatEventBeforeDone)?;
                     self.pending_devices.insert(
                         device.clone(),
                         DeviceInner {
@@ -185,7 +226,7 @@ impl EiEventConverter {
                     );
                 }
                 ei::seat::Event::Destroyed { serial } => {
-                    self.serial = serial;
+                    self.connection.update_serial(serial);
                     self.pending_seats.remove(&seat);
                     if let Some(seat) = self.seats.remove(&seat) {
                         self.queue_event(EiEvent::SeatRemoved(SeatRemoved { seat }));
@@ -197,21 +238,21 @@ impl EiEventConverter {
                     let device = self
                         .pending_devices
                         .get_mut(&device)
-                        .ok_or(Error::DeviceSetupEventAfterDone)?;
+                        .ok_or(EventError::DeviceSetupEventAfterDone)?;
                     device.name = Some(name);
                 }
                 ei::device::Event::DeviceType { device_type } => {
                     let device = self
                         .pending_devices
                         .get_mut(&device)
-                        .ok_or(Error::DeviceSetupEventAfterDone)?;
+                        .ok_or(EventError::DeviceSetupEventAfterDone)?;
                     device.device_type = Some(device_type);
                 }
                 ei::device::Event::Interface { object } => {
                     let device = self
                         .pending_devices
                         .get_mut(&device)
-                        .ok_or(Error::DeviceSetupEventAfterDone)?;
+                        .ok_or(EventError::DeviceSetupEventAfterDone)?;
                     device
                         .interfaces
                         .insert(object.interface().to_string(), object);
@@ -220,7 +261,7 @@ impl EiEventConverter {
                     let device = self
                         .pending_devices
                         .get_mut(&device)
-                        .ok_or(Error::DeviceSetupEventAfterDone)?;
+                        .ok_or(EventError::DeviceSetupEventAfterDone)?;
                     device.dimensions = Some((width, height));
                 }
                 ei::device::Event::Region {
@@ -233,7 +274,7 @@ impl EiEventConverter {
                     let device = self
                         .pending_devices
                         .get_mut(&device)
-                        .ok_or(Error::DeviceSetupEventAfterDone)?;
+                        .ok_or(EventError::DeviceSetupEventAfterDone)?;
                     device.regions.push(Region {
                         x: offset_x,
                         y: offset_y,
@@ -247,9 +288,9 @@ impl EiEventConverter {
                     let device = self
                         .pending_devices
                         .remove(&device)
-                        .ok_or(Error::DeviceSetupEventAfterDone)?;
+                        .ok_or(EventError::DeviceSetupEventAfterDone)?;
                     if device.device_type.is_none() {
-                        return Err(Error::NoDeviceType);
+                        return Err(EventError::NoDeviceType);
                     }
                     let device = Device(Arc::new(device));
                     self.devices.insert(device.0.device.clone(), device.clone());
@@ -259,33 +300,33 @@ impl EiEventConverter {
                     self.queue_event(EiEvent::DeviceAdded(DeviceAdded { device }));
                 }
                 ei::device::Event::Resumed { serial } => {
-                    self.serial = serial;
+                    self.connection.update_serial(serial);
                     let device = self
                         .devices
                         .get(&device)
-                        .ok_or(Error::DeviceEventBeforeDone)?;
+                        .ok_or(EventError::DeviceEventBeforeDone)?;
                     self.queue_event(EiEvent::DeviceResumed(DeviceResumed {
                         device: device.clone(),
                         serial,
                     }));
                 }
                 ei::device::Event::Paused { serial } => {
-                    self.serial = serial;
+                    self.connection.update_serial(serial);
                     let device = self
                         .devices
                         .get(&device)
-                        .ok_or(Error::DeviceEventBeforeDone)?;
+                        .ok_or(EventError::DeviceEventBeforeDone)?;
                     self.queue_event(EiEvent::DevicePaused(DevicePaused {
                         device: device.clone(),
                         serial,
                     }));
                 }
                 ei::device::Event::StartEmulating { serial, sequence } => {
-                    self.serial = serial;
+                    self.connection.update_serial(serial);
                     let device = self
                         .devices
                         .get(&device)
-                        .ok_or(Error::DeviceEventBeforeDone)?;
+                        .ok_or(EventError::DeviceEventBeforeDone)?;
                     self.queue_event(EiEvent::DeviceStartEmulating(DeviceStartEmulating {
                         device: device.clone(),
                         serial,
@@ -293,11 +334,11 @@ impl EiEventConverter {
                     }));
                 }
                 ei::device::Event::StopEmulating { serial } => {
-                    self.serial = serial;
+                    self.connection.update_serial(serial);
                     let device = self
                         .devices
                         .get(&device)
-                        .ok_or(Error::DeviceEventBeforeDone)?;
+                        .ok_or(EventError::DeviceEventBeforeDone)?;
                     self.queue_event(EiEvent::DeviceStopEmulating(DeviceStopEmulating {
                         device: device.clone(),
                         serial,
@@ -307,15 +348,15 @@ impl EiEventConverter {
                     let device = self
                         .pending_devices
                         .get_mut(&device)
-                        .ok_or(Error::DeviceEventBeforeDone)?;
+                        .ok_or(EventError::DeviceEventBeforeDone)?;
                     device.next_region_mapping_id = Some(mapping_id);
                 }
                 ei::device::Event::Frame { serial, timestamp } => {
-                    self.serial = serial;
+                    self.connection.update_serial(serial);
                     let device = self
                         .devices
                         .get(&device)
-                        .ok_or(Error::DeviceEventBeforeDone)?;
+                        .ok_or(EventError::DeviceEventBeforeDone)?;
                     self.queue_event(EiEvent::Frame(Frame {
                         device: device.clone(),
                         serial,
@@ -323,7 +364,7 @@ impl EiEventConverter {
                     }));
                 }
                 ei::device::Event::Destroyed { serial } => {
-                    self.serial = serial;
+                    self.connection.update_serial(serial);
                     self.pending_devices.remove(&device);
                     if let Some(device) = self.devices.remove(&device) {
                         self.queue_event(EiEvent::DeviceRemoved(DeviceRemoved { device }));
@@ -340,7 +381,7 @@ impl EiEventConverter {
                         .pending_devices
                         .values_mut()
                         .find(|i| i.interfaces.values().any(|j| j == &keyboard.0))
-                        .ok_or(Error::DeviceSetupEventAfterDone)?;
+                        .ok_or(EventError::DeviceSetupEventAfterDone)?;
                     device.keymap = Some(Keymap {
                         type_: keymap_type,
                         size,
@@ -351,7 +392,7 @@ impl EiEventConverter {
                     let device = self
                         .device_for_interface
                         .get(&keyboard.0)
-                        .ok_or(Error::DeviceEventBeforeDone)?;
+                        .ok_or(EventError::DeviceEventBeforeDone)?;
                     self.queue_event(EiEvent::KeyboardKey(KeyboardKey {
                         device: device.clone(),
                         time: 0,
@@ -366,11 +407,11 @@ impl EiEventConverter {
                     latched,
                     group,
                 } => {
-                    self.serial = serial;
+                    self.connection.update_serial(serial);
                     let device = self
                         .device_for_interface
                         .get(&keyboard.0)
-                        .ok_or(Error::DeviceEventBeforeDone)?;
+                        .ok_or(EventError::DeviceEventBeforeDone)?;
                     self.queue_event(EiEvent::KeyboardModifiers(KeyboardModifiers {
                         device: device.clone(),
                         serial,
@@ -381,7 +422,7 @@ impl EiEventConverter {
                     }));
                 }
                 ei::keyboard::Event::Destroyed { serial } => {
-                    self.serial = serial;
+                    self.connection.update_serial(serial);
                     // TODO does interface need to be removed from `Device`?
                     self.device_for_interface.remove(&keyboard.0);
                 }
@@ -390,7 +431,7 @@ impl EiEventConverter {
                 let device = self
                     .device_for_interface
                     .get(&pointer.0)
-                    .ok_or(Error::DeviceEventBeforeDone)?;
+                    .ok_or(EventError::DeviceEventBeforeDone)?;
                 match event {
                     ei::pointer::Event::MotionRelative { x, y } => {
                         self.queue_event(EiEvent::PointerMotion(PointerMotion {
@@ -401,7 +442,7 @@ impl EiEventConverter {
                         }));
                     }
                     ei::pointer::Event::Destroyed { serial } => {
-                        self.serial = serial;
+                        self.connection.update_serial(serial);
                         // TODO does interface need to be removed from `Device`?
                         self.device_for_interface.remove(&pointer.0);
                     }
@@ -411,7 +452,7 @@ impl EiEventConverter {
                 let device = self
                     .device_for_interface
                     .get(&pointer_absolute.0)
-                    .ok_or(Error::DeviceEventBeforeDone)?;
+                    .ok_or(EventError::DeviceEventBeforeDone)?;
                 match event {
                     ei::pointer_absolute::Event::MotionAbsolute { x, y } => {
                         self.queue_event(EiEvent::PointerMotionAbsolute(PointerMotionAbsolute {
@@ -422,7 +463,7 @@ impl EiEventConverter {
                         }));
                     }
                     ei::pointer_absolute::Event::Destroyed { serial } => {
-                        self.serial = serial;
+                        self.connection.update_serial(serial);
                         // TODO does interface need to be removed from `Device`?
                         self.device_for_interface.remove(&pointer_absolute.0);
                     }
@@ -432,7 +473,7 @@ impl EiEventConverter {
                 let device = self
                     .device_for_interface
                     .get(&scroll.0)
-                    .ok_or(Error::DeviceEventBeforeDone)?;
+                    .ok_or(EventError::DeviceEventBeforeDone)?;
                 match event {
                     ei::scroll::Event::Scroll { x, y } => {
                         self.queue_event(EiEvent::ScrollDelta(ScrollDelta {
@@ -468,7 +509,7 @@ impl EiEventConverter {
                         }
                     }
                     ei::scroll::Event::Destroyed { serial } => {
-                        self.serial = serial;
+                        self.connection.update_serial(serial);
                         // TODO does interface need to be removed from `Device`?
                         self.device_for_interface.remove(&scroll.0);
                     }
@@ -478,7 +519,7 @@ impl EiEventConverter {
                 let device = self
                     .device_for_interface
                     .get(&button.0)
-                    .ok_or(Error::DeviceEventBeforeDone)?;
+                    .ok_or(EventError::DeviceEventBeforeDone)?;
                 match event {
                     ei::button::Event::Button { button, state } => {
                         self.queue_event(EiEvent::Button(Button {
@@ -489,7 +530,7 @@ impl EiEventConverter {
                         }));
                     }
                     ei::button::Event::Destroyed { serial } => {
-                        self.serial = serial;
+                        self.connection.update_serial(serial);
                         // TODO does interface need to be removed from `Device`?
                         self.device_for_interface.remove(&button.0);
                     }
@@ -499,7 +540,7 @@ impl EiEventConverter {
                 let device = self
                     .device_for_interface
                     .get(&touchscreen.0)
-                    .ok_or(Error::DeviceEventBeforeDone)?;
+                    .ok_or(EventError::DeviceEventBeforeDone)?;
                 match event {
                     ei::touchscreen::Event::Down { touchid, x, y } => {
                         self.queue_event(EiEvent::TouchDown(TouchDown {
@@ -527,7 +568,7 @@ impl EiEventConverter {
                         }));
                     }
                     ei::touchscreen::Event::Destroyed { serial } => {
-                        self.serial = serial;
+                        self.connection.update_serial(serial);
                         // TODO does interface need to be removed from `Device`?
                         self.device_for_interface.remove(&touchscreen.0);
                     }
@@ -541,8 +582,12 @@ impl EiEventConverter {
         self.events.pop_front()
     }
 
-    pub fn serial(&self) -> u32 {
-        self.serial
+    pub fn add_callback_handler<F: FnOnce(u64) + 'static>(
+        &mut self,
+        callback: ei::Callback,
+        cb: F,
+    ) {
+        self.callbacks.insert(callback, Box::new(cb));
     }
 }
 
@@ -701,6 +746,10 @@ impl Device {
     pub fn interface<T: ei::Interface>(&self) -> Option<T> {
         self.0.interfaces.get(T::NAME)?.clone().downcast()
     }
+
+    pub fn has_capability(&self, capability: DeviceCapability) -> bool {
+        self.0.interfaces.contains_key(capability.name())
+    }
 }
 
 impl PartialEq for Device {
@@ -995,29 +1044,22 @@ impl_device_trait!(TouchDown; time);
 impl_device_trait!(TouchUp; time);
 impl_device_trait!(TouchMotion; time);
 
-#[derive(Debug)]
-pub enum EiConvertEventIteratorError {
-    Io(io::Error),
-    Parse(ParseError),
-    Event(crate::event::Error),
-}
-
 pub struct EiConvertEventIterator {
     context: ei::Context,
     converter: EiEventConverter,
 }
 
 impl EiConvertEventIterator {
-    pub fn new(context: ei::Context, serial: u32) -> Self {
+    fn new(context: ei::Context, handshake_resp: HandshakeResp) -> Self {
         Self {
+            converter: EiEventConverter::new(&context, handshake_resp),
             context,
-            converter: EiEventConverter::new(serial),
         }
     }
 }
 
 impl Iterator for EiConvertEventIterator {
-    type Item = Result<crate::event::EiEvent, EiConvertEventIteratorError>;
+    type Item = Result<crate::event::EiEvent, Error>;
 
     fn next(&mut self) -> Option<Self::Item> {
         loop {
@@ -1025,18 +1067,18 @@ impl Iterator for EiConvertEventIterator {
                 return Some(Ok(event));
             }
             if let Err(err) = util::poll_readable(&self.context) {
-                return Some(Err(EiConvertEventIteratorError::Io(err)));
+                return Some(Err(err.into()));
             }
             match self.context.read() {
                 Err(err) if err.kind() == io::ErrorKind::UnexpectedEof => return None,
-                Err(err) => return Some(Err(EiConvertEventIteratorError::Io(err))),
+                Err(err) => return Some(Err(err.into())),
                 Ok(_) => {}
             };
             while let Some(result) = self.context.pending_event() {
                 let request = match result {
                     PendingRequestResult::Request(request) => request,
                     PendingRequestResult::ParseError(parse_error) => {
-                        return Some(Err(EiConvertEventIteratorError::Parse(parse_error)))
+                        return Some(Err(parse_error.into()))
                     }
                     PendingRequestResult::InvalidObject(_invalid_object) => {
                         // Log?
@@ -1045,9 +1087,22 @@ impl Iterator for EiConvertEventIterator {
                 };
 
                 if let Err(err) = self.converter.handle_event(request) {
-                    return Some(Err(EiConvertEventIteratorError::Event(err)));
+                    return Some(Err(err.into()));
                 }
             }
         }
     }
 }
+
+impl ei::Context {
+    pub fn handshake_blocking(
+        &self,
+        name: &str,
+        context_type: ei::handshake::ContextType,
+    ) -> Result<(crate::event::Connection, EiConvertEventIterator), Error> {
+        let resp = crate::handshake::ei_handshake_blocking(self, name, context_type)?;
+        let iterator = EiConvertEventIterator::new(self.clone(), resp);
+        let connection = iterator.converter.connection().clone();
+        Ok((connection, iterator))
+    }
+}
