@@ -1,8 +1,6 @@
 #![forbid(unsafe_code)]
 
-// TODO error type?
 // TODO split up
-// Implement handshake
 
 use std::{env, path::PathBuf};
 
@@ -12,10 +10,11 @@ mod eiproto_ei;
 mod eiproto_eis;
 mod eiproto_enum;
 pub mod eis;
+mod error;
+pub use error::Error;
 pub mod event; // XXX reorganize?
 pub mod handshake; // XXX ^
 mod object;
-#[doc(hidden)] // TODO
 pub mod request;
 pub use object::Object;
 mod util;
@@ -26,7 +25,7 @@ pub use wire::Interface;
 pub use wire::ParseError;
 
 #[cfg(feature = "calloop")]
-#[doc(hidden)] // TODO
+//#[doc(hidden)] // TODO
 pub mod calloop;
 #[cfg(feature = "tokio")]
 pub mod tokio;
