@@ -2,75 +2,154 @@
 
 // TODO: rename/reorganize things; doc comments on public types/methods
 
-use crate::{eis, Object, ParseError};
+use crate::{eis, handshake::EisHandshakeResp, wire::Interface, Error, Object};
 use std::{
     collections::{HashMap, VecDeque},
-    fmt, io,
-    sync::Arc,
+    fmt,
+    sync::{Arc, Mutex, Weak},
 };
 
 pub use crate::event::DeviceCapability;
 
 #[derive(Debug)]
-pub enum Error {
-    UnexpectedHandshakeEvent,
-    UnrecognizedSeat,
-    UnrecognizedDevice,
-    InvalidCallbackVersion,
-    Parse(ParseError),
-    Io(io::Error),
+struct ConnectionInner {
+    context: eis::Context,
+    handshake_resp: EisHandshakeResp,
+    seats: Mutex<HashMap<eis::Seat, Seat>>,
+    devices: Mutex<HashMap<eis::Device, Device>>,
+    device_for_interface: Mutex<HashMap<Object, Device>>,
+    last_serial: Mutex<u32>,
 }
 
-impl fmt::Display for Error {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        match self {
-            Self::UnexpectedHandshakeEvent => write!(f, "unexpected handshake event"),
-            Self::UnrecognizedSeat => write!(f, "unrecognized seat"),
-            Self::UnrecognizedDevice => write!(f, "unrecognized device"),
-            Self::InvalidCallbackVersion => write!(f, "invalid callback version"),
-            Self::Io(err) => write!(f, "IO error: {}", err),
-            Self::Parse(err) => write!(f, "parse error: {}", err),
-        }
+#[derive(Clone, Debug)]
+pub struct Connection(Arc<ConnectionInner>);
+
+impl Connection {
+    pub fn connection(&self) -> &eis::Connection {
+        &self.0.handshake_resp.connection
     }
-}
 
-impl std::error::Error for Error {}
-
-// need way to add seat/device?
-pub struct EisRequestConverter {
-    connection: eis::Connection,
-    requests: VecDeque<EisRequest>,
-    pending_requests: VecDeque<EisRequest>,
-    seats: HashMap<eis::Seat, Seat>,
-    devices: HashMap<eis::Device, Device>,
-    device_for_interface: HashMap<Object, Device>,
-    last_serial: u32,
-}
-
-impl EisRequestConverter {
-    pub fn new(connection: &eis::Connection, initial_serial: u32) -> Self {
-        Self {
-            connection: connection.clone(),
-            last_serial: initial_serial,
-            requests: VecDeque::new(),
-            pending_requests: VecDeque::new(),
-            seats: HashMap::new(),
-            devices: HashMap::new(),
-            device_for_interface: HashMap::new(),
+    pub fn disconnected(&self, reason: eis::connection::DisconnectReason, explanation: &str) {
+        let seats = self
+            .0
+            .seats
+            .lock()
+            .unwrap()
+            .values()
+            .cloned()
+            .collect::<Vec<_>>();
+        for seat in seats {
+            seat.remove();
         }
+        self.connection()
+            .disconnected(self.last_serial(), reason, explanation);
+    }
+
+    pub fn flush(&self) -> rustix::io::Result<()> {
+        self.0.context.flush()
+    }
+
+    pub fn context_type(&self) -> eis::handshake::ContextType {
+        self.0.handshake_resp.context_type
+    }
+
+    pub fn name(&self) -> Option<&str> {
+        self.0.handshake_resp.name.as_deref()
+    }
+
+    // Use type instead of string?
+    pub fn has_interface(&self, interface: &str) -> bool {
+        self.0
+            .handshake_resp
+            .negotiated_interfaces
+            .contains_key(interface)
+    }
+
+    pub fn interface_version(&self, interface: &str) -> Option<u32> {
+        self.0
+            .handshake_resp
+            .negotiated_interfaces
+            .get(interface)
+            .copied()
     }
 
     pub fn last_serial(&self) -> u32 {
-        self.last_serial
+        *self.0.last_serial.lock().unwrap()
     }
 
-    pub fn next_serial(&mut self) -> u32 {
-        self.last_serial += 1;
-        self.last_serial
+    pub fn with_next_serial<T, F: FnOnce(u32) -> T>(&self, cb: F) -> T {
+        let mut last_serial = self.0.last_serial.lock().unwrap();
+        let serial = last_serial.wrapping_add(1);
+        let res = cb(serial);
+        *last_serial = serial;
+        res
+    }
+
+    fn device_for_interface<T: DeviceInterface>(&mut self, interface: &T) -> Option<Device> {
+        self.0
+            .device_for_interface
+            .lock()
+            .unwrap()
+            .get(interface.as_object())
+            .cloned()
+    }
+
+    pub fn add_seat(&self, name: Option<&str>, capabilities: &[DeviceCapability]) -> Seat {
+        let seat = self.connection().seat(1);
+        if let Some(name) = name {
+            seat.name(name);
+        }
+        for capability in capabilities {
+            // TODO only send negotiated interfaces
+            seat.capability(2 << *capability as u64, capability.name());
+        }
+        seat.done();
+        let seat = Seat(Arc::new(SeatInner {
+            seat,
+            name: name.map(|x| x.to_owned()),
+            handle: Arc::downgrade(&self.0),
+        }));
+        self.0
+            .seats
+            .lock()
+            .unwrap()
+            .insert(seat.0.seat.clone(), seat.clone());
+        seat
     }
 }
 
+// need way to add seat/device?
+#[derive(Debug)]
+pub struct EisRequestConverter {
+    requests: VecDeque<EisRequest>,
+    pending_requests: VecDeque<EisRequest>,
+    handle: Connection,
+}
+
 impl EisRequestConverter {
+    pub fn new(
+        context: &eis::Context,
+        handshake_resp: EisHandshakeResp,
+        initial_serial: u32,
+    ) -> Self {
+        Self {
+            requests: VecDeque::new(),
+            pending_requests: VecDeque::new(),
+            handle: Connection(Arc::new(ConnectionInner {
+                context: context.clone(),
+                handshake_resp,
+                seats: Default::default(),
+                devices: Default::default(),
+                device_for_interface: Default::default(),
+                last_serial: Mutex::new(initial_serial),
+            })),
+        }
+    }
+
+    pub fn handle(&self) -> &Connection {
+        &self.handle
+    }
+
     // Based on behavior of `eis_queue_request` in libeis
     fn queue_request(&mut self, mut request: EisRequest) {
         if request.time_mut().is_some() {
@@ -94,33 +173,254 @@ impl EisRequestConverter {
         self.requests.pop_front()
     }
 
-    pub fn add_seat(&mut self, name: Option<&str>, capabilities: &[DeviceCapability]) -> Seat {
-        let seat = self.connection.seat(1);
-        if let Some(name) = name {
-            seat.name(name);
+    pub fn handle_request(&mut self, request: eis::Request) -> Result<(), Error> {
+        match request {
+            eis::Request::Handshake(_handshake, _request) => {
+                return Err(Error::UnexpectedHandshakeEvent);
+            }
+            eis::Request::Connection(_connection, request) => match request {
+                eis::connection::Request::Sync { callback } => {
+                    if callback.version() != 1 {
+                        return Err(Error::InvalidInterfaceVersion(
+                            "ei_callback",
+                            callback.version(),
+                        ));
+                    }
+                    callback.done(0);
+                    if let Some(backend) = callback.0.backend() {
+                        // XXX Error?
+                        let _ = backend.flush();
+                    }
+                }
+                eis::connection::Request::Disconnect => {
+                    self.queue_request(EisRequest::Disconnect);
+                }
+            },
+            eis::Request::Callback(_callback, request) => match request {},
+            eis::Request::Pingpong(_ping_pong, request) => match request {
+                eis::pingpong::Request::Done { callback_data: _ } => {
+                    // TODO
+                }
+            },
+            eis::Request::Seat(seat, request) => match request {
+                eis::seat::Request::Release => {
+                    self.handle
+                        .with_next_serial(|serial| seat.destroyed(serial));
+                }
+                eis::seat::Request::Bind { capabilities } => {
+                    let Some(seat) = self.handle.0.seats.lock().unwrap().get(&seat).cloned() else {
+                        return Ok(());
+                    };
+                    self.queue_request(EisRequest::Bind(Bind { seat, capabilities }));
+                }
+            },
+            eis::Request::Device(device, request) => {
+                let Some(device) = self.handle.0.devices.lock().unwrap().get(&device).cloned()
+                else {
+                    return Ok(());
+                };
+                match request {
+                    eis::device::Request::Release => {}
+                    eis::device::Request::StartEmulating {
+                        last_serial,
+                        sequence,
+                    } => {
+                        self.queue_request(EisRequest::DeviceStartEmulating(
+                            DeviceStartEmulating {
+                                device,
+                                last_serial,
+                                sequence,
+                            },
+                        ));
+                    }
+                    eis::device::Request::StopEmulating { last_serial } => {
+                        self.queue_request(EisRequest::DeviceStopEmulating(DeviceStopEmulating {
+                            device,
+                            last_serial,
+                        }));
+                    }
+                    eis::device::Request::Frame {
+                        last_serial,
+                        timestamp,
+                    } => {
+                        self.queue_request(EisRequest::Frame(Frame {
+                            device,
+                            last_serial,
+                            time: timestamp,
+                        }));
+                    }
+                }
+            }
+            eis::Request::Keyboard(keyboard, request) => {
+                let Some(device) = self.handle.device_for_interface(&keyboard) else {
+                    return Ok(());
+                };
+                match request {
+                    eis::keyboard::Request::Release => {}
+                    eis::keyboard::Request::Key { key, state } => {
+                        self.queue_request(EisRequest::KeyboardKey(KeyboardKey {
+                            device,
+                            key,
+                            state,
+                            time: 0,
+                        }));
+                    }
+                }
+            }
+            eis::Request::Pointer(pointer, request) => {
+                let Some(device) = self.handle.device_for_interface(&pointer) else {
+                    return Ok(());
+                };
+                match request {
+                    eis::pointer::Request::Release => {}
+                    eis::pointer::Request::MotionRelative { x, y } => {
+                        self.queue_request(EisRequest::PointerMotion(PointerMotion {
+                            device,
+                            dx: x,
+                            dy: y,
+                            time: 0,
+                        }));
+                    }
+                }
+            }
+            eis::Request::PointerAbsolute(pointer_absolute, request) => {
+                let Some(device) = self.handle.device_for_interface(&pointer_absolute) else {
+                    return Ok(());
+                };
+                match request {
+                    eis::pointer_absolute::Request::Release => {}
+                    eis::pointer_absolute::Request::MotionAbsolute { x, y } => {
+                        self.queue_request(EisRequest::PointerMotionAbsolute(
+                            PointerMotionAbsolute {
+                                device,
+                                dx_absolute: x,
+                                dy_absolute: y,
+                                time: 0,
+                            },
+                        ));
+                    }
+                }
+            }
+            eis::Request::Scroll(scroll, request) => {
+                let Some(device) = self.handle.device_for_interface(&scroll) else {
+                    return Ok(());
+                };
+                match request {
+                    eis::scroll::Request::Release => {}
+                    eis::scroll::Request::Scroll { x, y } => {
+                        self.queue_request(EisRequest::ScrollDelta(ScrollDelta {
+                            device,
+                            dx: x,
+                            dy: y,
+                            time: 0,
+                        }));
+                    }
+                    eis::scroll::Request::ScrollDiscrete { x, y } => {
+                        self.queue_request(EisRequest::ScrollDiscrete(ScrollDiscrete {
+                            device,
+                            discrete_dx: x,
+                            discrete_dy: y,
+                            time: 0,
+                        }));
+                    }
+                    eis::scroll::Request::ScrollStop { x, y, is_cancel } => {
+                        if is_cancel != 0 {
+                            self.queue_request(EisRequest::ScrollCancel(ScrollCancel {
+                                device,
+                                time: 0,
+                                x: x != 0,
+                                y: y != 0,
+                            }));
+                        } else {
+                            self.queue_request(EisRequest::ScrollStop(ScrollStop {
+                                device,
+                                time: 0,
+                                x: x != 0,
+                                y: y != 0,
+                            }));
+                        }
+                    }
+                }
+            }
+            eis::Request::Button(button, request) => {
+                let Some(device) = self.handle.device_for_interface(&button) else {
+                    return Ok(());
+                };
+                match request {
+                    eis::button::Request::Release => {}
+                    eis::button::Request::Button { button, state } => {
+                        self.queue_request(EisRequest::Button(Button {
+                            device,
+                            button,
+                            state,
+                            time: 0,
+                        }));
+                    }
+                }
+            }
+            eis::Request::Touchscreen(touchscreen, request) => {
+                let Some(device) = self.handle.device_for_interface(&touchscreen) else {
+                    return Ok(());
+                };
+                match request {
+                    eis::touchscreen::Request::Release => {}
+                    eis::touchscreen::Request::Down { touchid, x, y } => {
+                        self.queue_request(EisRequest::TouchDown(TouchDown {
+                            device,
+                            touch_id: touchid,
+                            x,
+                            y,
+                            time: 0,
+                        }));
+                    }
+                    eis::touchscreen::Request::Motion { touchid, x, y } => {
+                        self.queue_request(EisRequest::TouchMotion(TouchMotion {
+                            device,
+                            touch_id: touchid,
+                            x,
+                            y,
+                            time: 0,
+                        }));
+                    }
+                    eis::touchscreen::Request::Up { touchid } => {
+                        self.queue_request(EisRequest::TouchUp(TouchUp {
+                            device,
+                            touch_id: touchid,
+                            time: 0,
+                        }));
+                    }
+                }
+            }
         }
-        for capability in capabilities {
-            // TODO only send negotiated interfaces
-            seat.capability(2 << *capability as u64, capability.name());
-        }
-        seat.done();
-        let seat = Seat(Arc::new(SeatInner {
-            seat,
-            name: name.map(|x| x.to_owned()),
-        }));
-        self.seats.insert(seat.0.seat.clone(), seat.clone());
-        seat
+        Ok(())
+    }
+}
+
+struct SeatInner {
+    seat: eis::Seat,
+    name: Option<String>,
+    //capabilities: HashMap<String, u64>,
+    handle: Weak<ConnectionInner>,
+}
+
+#[derive(Clone)]
+pub struct Seat(Arc<SeatInner>);
+
+impl Seat {
+    pub fn eis_seat(&self) -> &eis::Seat {
+        &self.0.seat
     }
 
     // builder pattern?
     pub fn add_device(
-        &mut self,
-        seat: &Seat,
+        &self,
         name: Option<&str>,
         device_type: eis::device::DeviceType,
         capabilities: &[DeviceCapability],
+        // TODO: better solution; keymap, etc.
+        before_done_cb: impl for<'a> FnOnce(&'a Device),
     ) -> Device {
-        let device = seat.0.seat.device(1);
+        let device = self.0.seat.device(1);
         if let Some(name) = name {
             device.name(name);
         }
@@ -142,259 +442,55 @@ impl EisRequestConverter {
             };
             interfaces.insert(object.interface().to_string(), object);
         }
-        device.done();
 
         let device = Device(Arc::new(DeviceInner {
             device,
-            seat: seat.clone(),
+            seat: self.clone(),
             name: name.map(|x| x.to_string()),
             interfaces,
+            handle: self.0.handle.clone(),
         }));
-        for interface in device.0.interfaces.values() {
-            self.device_for_interface
-                .insert(interface.clone(), device.clone());
+        if let Some(handle) = self.0.handle.upgrade().map(Connection) {
+            for interface in device.0.interfaces.values() {
+                handle
+                    .0
+                    .device_for_interface
+                    .lock()
+                    .unwrap()
+                    .insert(interface.clone(), device.clone());
+            }
+            handle
+                .0
+                .devices
+                .lock()
+                .unwrap()
+                .insert(device.0.device.clone(), device.clone());
         }
-        self.devices.insert(device.0.device.clone(), device.clone());
+
+        before_done_cb(&device);
+        device.device().done();
+
         device
     }
 
-    pub fn handle_request(&mut self, request: eis::Request) -> Result<(), Error> {
-        match request {
-            eis::Request::Handshake(_handshake, _request) => {
-                return Err(Error::UnexpectedHandshakeEvent);
-            }
-            eis::Request::Connection(_connection, request) => match request {
-                eis::connection::Request::Sync { callback } => {
-                    if callback.version() != 1 {
-                        return Err(Error::InvalidCallbackVersion);
-                    }
-                    callback.done(0);
-                    if let Some(backend) = callback.0.backend() {
-                        // XXX Error?
-                        let _ = backend.flush();
-                    }
-                }
-                eis::connection::Request::Disconnect => {
-                    self.queue_request(EisRequest::Disconnect);
-                }
-            },
-            eis::Request::Callback(_callback, request) => match request {},
-            eis::Request::Pingpong(_ping_pong, request) => match request {
-                eis::pingpong::Request::Done { callback_data: _ } => {
-                    // TODO
-                }
-            },
-            eis::Request::Seat(seat, request) => match request {
-                eis::seat::Request::Release => {
-                    // XXX
-                    let serial = self.next_serial();
-                    seat.destroyed(serial);
-                }
-                eis::seat::Request::Bind { capabilities } => {
-                    let seat = self.seats.get(&seat).ok_or(Error::UnrecognizedSeat)?;
-                    self.queue_request(EisRequest::Bind(Bind {
-                        seat: seat.clone(),
-                        capabilities,
-                    }));
-                }
-            },
-            eis::Request::Device(device, request) => {
-                let device = self.devices.get(&device).ok_or(Error::UnrecognizedDevice)?;
-                match request {
-                    eis::device::Request::Release => {}
-                    eis::device::Request::StartEmulating {
-                        last_serial,
-                        sequence,
-                    } => {
-                        self.queue_request(EisRequest::DeviceStartEmulating(
-                            DeviceStartEmulating {
-                                device: device.clone(),
-                                last_serial,
-                                sequence,
-                            },
-                        ));
-                    }
-                    eis::device::Request::StopEmulating { last_serial } => {
-                        self.queue_request(EisRequest::DeviceStopEmulating(DeviceStopEmulating {
-                            device: device.clone(),
-                            last_serial,
-                        }));
-                    }
-                    eis::device::Request::Frame {
-                        last_serial,
-                        timestamp,
-                    } => {
-                        self.queue_request(EisRequest::Frame(Frame {
-                            device: device.clone(),
-                            last_serial,
-                            time: timestamp,
-                        }));
-                    }
-                }
-            }
-            eis::Request::Keyboard(keyboard, request) => {
-                let device = self
-                    .device_for_interface
-                    .get(&keyboard.0)
-                    .ok_or(Error::UnrecognizedDevice)?;
-                match request {
-                    eis::keyboard::Request::Release => {}
-                    eis::keyboard::Request::Key { key, state } => {
-                        self.queue_request(EisRequest::KeyboardKey(KeyboardKey {
-                            device: device.clone(),
-                            key,
-                            state,
-                            time: 0,
-                        }));
-                    }
-                }
-            }
-            eis::Request::Pointer(pointer, request) => {
-                let device = self
-                    .device_for_interface
-                    .get(&pointer.0)
-                    .ok_or(Error::UnrecognizedDevice)?;
-                match request {
-                    eis::pointer::Request::Release => {}
-                    eis::pointer::Request::MotionRelative { x, y } => {
-                        self.queue_request(EisRequest::PointerMotion(PointerMotion {
-                            device: device.clone(),
-                            dx: x,
-                            dy: y,
-                            time: 0,
-                        }));
-                    }
-                }
-            }
-            eis::Request::PointerAbsolute(pointer_absolute, request) => {
-                let device = self
-                    .device_for_interface
-                    .get(&pointer_absolute.0)
-                    .ok_or(Error::UnrecognizedDevice)?;
-                match request {
-                    eis::pointer_absolute::Request::Release => {}
-                    eis::pointer_absolute::Request::MotionAbsolute { x, y } => {
-                        self.queue_request(EisRequest::PointerMotionAbsolute(
-                            PointerMotionAbsolute {
-                                device: device.clone(),
-                                dx_absolute: x,
-                                dy_absolute: y,
-                                time: 0,
-                            },
-                        ));
-                    }
-                }
-            }
-            eis::Request::Scroll(scroll, request) => {
-                let device = self
-                    .device_for_interface
-                    .get(&scroll.0)
-                    .ok_or(Error::UnrecognizedDevice)?;
-                match request {
-                    eis::scroll::Request::Release => {}
-                    eis::scroll::Request::Scroll { x, y } => {
-                        self.queue_request(EisRequest::ScrollDelta(ScrollDelta {
-                            device: device.clone(),
-                            dx: x,
-                            dy: y,
-                            time: 0,
-                        }));
-                    }
-                    eis::scroll::Request::ScrollDiscrete { x, y } => {
-                        self.queue_request(EisRequest::ScrollDiscrete(ScrollDiscrete {
-                            device: device.clone(),
-                            discrete_dx: x,
-                            discrete_dy: y,
-                            time: 0,
-                        }));
-                    }
-                    eis::scroll::Request::ScrollStop { x, y, is_cancel } => {
-                        if is_cancel != 0 {
-                            self.queue_request(EisRequest::ScrollCancel(ScrollCancel {
-                                device: device.clone(),
-                                time: 0,
-                                x: x != 0,
-                                y: y != 0,
-                            }));
-                        } else {
-                            self.queue_request(EisRequest::ScrollStop(ScrollStop {
-                                device: device.clone(),
-                                time: 0,
-                                x: x != 0,
-                                y: y != 0,
-                            }));
-                        }
-                    }
-                }
-            }
-            eis::Request::Button(button, request) => {
-                let device = self
-                    .device_for_interface
-                    .get(&button.0)
-                    .ok_or(Error::UnrecognizedDevice)?;
-                match request {
-                    eis::button::Request::Release => {}
-                    eis::button::Request::Button { button, state } => {
-                        self.queue_request(EisRequest::Button(Button {
-                            device: device.clone(),
-                            button,
-                            state,
-                            time: 0,
-                        }));
-                    }
-                }
-            }
-            eis::Request::Touchscreen(touchscreen, request) => {
-                let device = self
-                    .device_for_interface
-                    .get(&touchscreen.0)
-                    .ok_or(Error::UnrecognizedDevice)?;
-                match request {
-                    eis::touchscreen::Request::Release => {}
-                    eis::touchscreen::Request::Down { touchid, x, y } => {
-                        self.queue_request(EisRequest::TouchDown(TouchDown {
-                            device: device.clone(),
-                            touch_id: touchid,
-                            x,
-                            y,
-                            time: 0,
-                        }));
-                    }
-                    eis::touchscreen::Request::Motion { touchid, x, y } => {
-                        self.queue_request(EisRequest::TouchMotion(TouchMotion {
-                            device: device.clone(),
-                            touch_id: touchid,
-                            x,
-                            y,
-                            time: 0,
-                        }));
-                    }
-                    eis::touchscreen::Request::Up { touchid } => {
-                        self.queue_request(EisRequest::TouchUp(TouchUp {
-                            device: device.clone(),
-                            touch_id: touchid,
-                            time: 0,
-                        }));
-                    }
-                }
+    pub fn remove(&self) {
+        if let Some(handle) = self.0.handle.upgrade().map(Connection) {
+            let devices = handle
+                .0
+                .devices
+                .lock()
+                .unwrap()
+                .values()
+                .filter(|device| &device.0.seat == self)
+                .cloned()
+                .collect::<Vec<_>>();
+            for device in devices {
+                device.remove();
             }
+
+            handle.with_next_serial(|serial| self.0.seat.destroyed(serial));
+            handle.0.seats.lock().unwrap().remove(&self.0.seat);
         }
-        Ok(())
-    }
-}
-
-struct SeatInner {
-    seat: eis::Seat,
-    name: Option<String>,
-    //capabilities: HashMap<String, u64>,
-}
-
-#[derive(Clone)]
-pub struct Seat(Arc<SeatInner>);
-
-impl Seat {
-    pub fn eis_seat(&self) -> &eis::Seat {
-        &self.0.seat
     }
 }
 
@@ -422,11 +518,46 @@ impl std::hash::Hash for Seat {
     }
 }
 
+pub trait DeviceInterface: eis::Interface {}
+
+macro_rules! impl_device_interface {
+    ($ty:ty) => {
+        impl DeviceInterface for $ty {}
+    };
+}
+impl_device_interface!(eis::Pointer);
+impl_device_interface!(eis::PointerAbsolute);
+impl_device_interface!(eis::Scroll);
+impl_device_interface!(eis::Button);
+impl_device_interface!(eis::Keyboard);
+impl_device_interface!(eis::Touchscreen);
+
+fn destroy_interface(object: crate::Object, serial: u32) {
+    match object.interface() {
+        eis::Pointer::NAME => object
+            .downcast_unchecked::<eis::Pointer>()
+            .destroyed(serial),
+        eis::PointerAbsolute::NAME => object
+            .downcast_unchecked::<eis::PointerAbsolute>()
+            .destroyed(serial),
+        eis::Scroll::NAME => object.downcast_unchecked::<eis::Scroll>().destroyed(serial),
+        eis::Button::NAME => object.downcast_unchecked::<eis::Button>().destroyed(serial),
+        eis::Keyboard::NAME => object
+            .downcast_unchecked::<eis::Keyboard>()
+            .destroyed(serial),
+        eis::Touchscreen::NAME => object
+            .downcast_unchecked::<eis::Touchscreen>()
+            .destroyed(serial),
+        _ => unreachable!(),
+    }
+}
+
 struct DeviceInner {
     device: eis::Device,
     seat: Seat,
     name: Option<String>,
     interfaces: HashMap<String, crate::Object>,
+    handle: Weak<ConnectionInner>,
 }
 
 #[derive(Clone)]
@@ -454,6 +585,62 @@ impl Device {
     pub fn name(&self) -> Option<&str> {
         self.0.name.as_deref()
     }
+
+    pub fn interface<T: DeviceInterface>(&self) -> Option<T> {
+        self.0.interfaces.get(T::NAME)?.clone().downcast()
+    }
+
+    pub fn has_capability(&self, capability: DeviceCapability) -> bool {
+        self.0.interfaces.contains_key(capability.name())
+    }
+
+    pub fn remove(&self) {
+        if let Some(handle) = self.0.handle.upgrade().map(Connection) {
+            for interface in self.0.interfaces.values() {
+                handle
+                    .0
+                    .device_for_interface
+                    .lock()
+                    .unwrap()
+                    .remove(interface);
+                handle.with_next_serial(|serial| destroy_interface(interface.clone(), serial));
+            }
+
+            handle.with_next_serial(|serial| self.0.device.destroyed(serial));
+            handle.0.devices.lock().unwrap().remove(&self.0.device);
+        }
+    }
+
+    pub fn resumed(&self) {
+        if let Some(handle) = self.0.handle.upgrade().map(Connection) {
+            handle.with_next_serial(|serial| self.device().resumed(serial))
+        }
+    }
+
+    pub fn paused(&self) {
+        if let Some(handle) = self.0.handle.upgrade().map(Connection) {
+            handle.with_next_serial(|serial| self.device().paused(serial))
+        }
+    }
+
+    // TODO: statically restrict to receiver context?
+    pub fn start_emulating(&self, sequence: u32) {
+        if let Some(handle) = self.0.handle.upgrade().map(Connection) {
+            handle.with_next_serial(|serial| self.device().start_emulating(serial, sequence))
+        }
+    }
+
+    pub fn stop_emulating(&self) {
+        if let Some(handle) = self.0.handle.upgrade().map(Connection) {
+            handle.with_next_serial(|serial| self.device().stop_emulating(serial))
+        }
+    }
+
+    pub fn frame(&self, time: u64) {
+        if let Some(handle) = self.0.handle.upgrade().map(Connection) {
+            handle.with_next_serial(|serial| self.device().frame(serial, time))
+        }
+    }
 }
 
 impl PartialEq for Device {
